@@ -1,4 +1,4 @@
-// Copyright (c) 2021 Uber Technologies, Inc.
+// Copyright (c) 2024 Uber Technologies, Inc.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -61,7 +61,7 @@ const (
 	// DefaultMaxQueueSize is the default M3 reporter queue size.
 	DefaultMaxQueueSize = 4096
 	// DefaultMaxPacketSize is the default M3 reporter max packet size.
-	DefaultMaxPacketSize = int32(1440)
+	DefaultMaxPacketSize = int32(32768)
 	// DefaultHistogramBucketIDName is the default histogram bucket ID tag name
 	DefaultHistogramBucketIDName = "bucketid"
 	// DefaultHistogramBucketName is the default histogram bucket name tag name
@@ -133,6 +133,7 @@ type reporter struct {
 	numMetricsCounter     tally.CachedCount
 	numWriteErrors        atomic.Int64
 	numWriteErrorsCounter tally.CachedCount
+	numTagCacheCounter    tally.CachedCount
 }
 
 // Options is a set of options for the M3 reporter.
@@ -148,6 +149,7 @@ type Options struct {
 	HistogramBucketIDName       string
 	HistogramBucketName         string
 	HistogramBucketTagPrecision uint
+	InternalTags                map[string]string
 }
 
 // NewReporter creates a new M3 reporter.
@@ -285,13 +287,20 @@ func NewReporter(opts Options) (Reporter, error) {
 	}
 
 	internalTags := map[string]string{
-		"version": tally.Version,
+		"version":  tally.Version,
+		"host":     tally.DefaultTagRedactValue,
+		"instance": tally.DefaultTagRedactValue,
 	}
+
+	for k, v := range opts.InternalTags {
+		internalTags[k] = v
+	}
+
 	r.batchSizeHistogram = r.AllocateHistogram("tally.internal.batch-size", internalTags, buckets)
 	r.numBatchesCounter = r.AllocateCounter("tally.internal.num-batches", internalTags)
 	r.numMetricsCounter = r.AllocateCounter("tally.internal.num-metrics", internalTags)
 	r.numWriteErrorsCounter = r.AllocateCounter("tally.internal.num-write-errors", internalTags)
-
+	r.numTagCacheCounter = r.AllocateCounter("tally.internal.num-tag-cache", internalTags)
 	r.wg.Add(1)
 	go func() {
 		defer r.wg.Done()
@@ -689,12 +698,19 @@ func (r *reporter) reportInternalMetrics() {
 	r.numBatchesCounter.ReportCount(batches)
 	r.numMetricsCounter.ReportCount(metrics)
 	r.numWriteErrorsCounter.ReportCount(writeErrors)
+	r.numTagCacheCounter.ReportCount(int64(r.tagCache.Len()))
 }
 
 func (r *reporter) timeLoop() {
+	t := time.NewTicker(_timeResolution)
+	defer t.Stop()
 	for !r.done.Load() {
 		r.now.Store(time.Now().UnixNano())
-		time.Sleep(_timeResolution)
+		select {
+		case <-t.C:
+		case <-r.donech:
+			return
+		}
 	}
 }
 
