@@ -1,4 +1,4 @@
-// Copyright (c) 2021 Uber Technologies, Inc.
+// Copyright (c) 2023 Uber Technologies, Inc.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -33,8 +33,13 @@ import (
 
 	"github.com/stretchr/testify/assert"
 	"github.com/stretchr/testify/require"
+	"go.uber.org/goleak"
 )
 
+func TestMain(m *testing.M) {
+	goleak.VerifyTestMain(m)
+}
+
 var (
 	// alphanumericSanitizerOpts is the options to create a sanitizer which uses
 	// the alphanumeric SanitizeFn.
@@ -96,13 +101,13 @@ func newTestHistogramValue() *testHistogramValue {
 }
 
 type testStatsReporter struct {
+	mtx sync.Mutex
+
 	cg sync.WaitGroup
 	gg sync.WaitGroup
 	tg sync.WaitGroup
 	hg sync.WaitGroup
 
-	scope Scope
-
 	counters   map[string]*testIntValue
 	gauges     map[string]*testFloatValue
 	timers     map[string]*testIntValue
@@ -122,6 +127,9 @@ func newTestStatsReporter() *testStatsReporter {
 }
 
 func (r *testStatsReporter) getCounters() map[string]*testIntValue {
+	r.mtx.Lock()
+	defer r.mtx.Unlock()
+
 	dst := make(map[string]*testIntValue, len(r.counters))
 	for k, v := range r.counters {
 		var (
@@ -139,6 +147,9 @@ func (r *testStatsReporter) getCounters() map[string]*testIntValue {
 }
 
 func (r *testStatsReporter) getGauges() map[string]*testFloatValue {
+	r.mtx.Lock()
+	defer r.mtx.Unlock()
+
 	dst := make(map[string]*testFloatValue, len(r.gauges))
 	for k, v := range r.gauges {
 		var (
@@ -156,6 +167,9 @@ func (r *testStatsReporter) getGauges() map[string]*testFloatValue {
 }
 
 func (r *testStatsReporter) getTimers() map[string]*testIntValue {
+	r.mtx.Lock()
+	defer r.mtx.Unlock()
+
 	dst := make(map[string]*testIntValue, len(r.timers))
 	for k, v := range r.timers {
 		var (
@@ -173,6 +187,9 @@ func (r *testStatsReporter) getTimers() map[string]*testIntValue {
 }
 
 func (r *testStatsReporter) getHistograms() map[string]*testHistogramValue {
+	r.mtx.Lock()
+	defer r.mtx.Unlock()
+
 	dst := make(map[string]*testHistogramValue, len(r.histograms))
 	for k, v := range r.histograms {
 		var (
@@ -199,6 +216,9 @@ func (r *testStatsReporter) WaitAll() {
 func (r *testStatsReporter) AllocateCounter(
 	name string, tags map[string]string,
 ) CachedCount {
+	r.mtx.Lock()
+	defer r.mtx.Unlock()
+
 	counter := &testIntValue{
 		val:      0,
 		tags:     tags,
@@ -209,6 +229,9 @@ func (r *testStatsReporter) AllocateCounter(
 }
 
 func (r *testStatsReporter) ReportCounter(name string, tags map[string]string, value int64) {
+	r.mtx.Lock()
+	defer r.mtx.Unlock()
+
 	r.counters[name] = &testIntValue{
 		val:  value,
 		tags: tags,
@@ -219,6 +242,9 @@ func (r *testStatsReporter) ReportCounter(name string, tags map[string]string, v
 func (r *testStatsReporter) AllocateGauge(
 	name string, tags map[string]string,
 ) CachedGauge {
+	r.mtx.Lock()
+	defer r.mtx.Unlock()
+
 	gauge := &testFloatValue{
 		val:      0,
 		tags:     tags,
@@ -229,6 +255,9 @@ func (r *testStatsReporter) AllocateGauge(
 }
 
 func (r *testStatsReporter) ReportGauge(name string, tags map[string]string, value float64) {
+	r.mtx.Lock()
+	defer r.mtx.Unlock()
+
 	r.gauges[name] = &testFloatValue{
 		val:  value,
 		tags: tags,
@@ -239,6 +268,9 @@ func (r *testStatsReporter) ReportGauge(name string, tags map[string]string, val
 func (r *testStatsReporter) AllocateTimer(
 	name string, tags map[string]string,
 ) CachedTimer {
+	r.mtx.Lock()
+	defer r.mtx.Unlock()
+
 	timer := &testIntValue{
 		val:      0,
 		tags:     tags,
@@ -249,6 +281,9 @@ func (r *testStatsReporter) AllocateTimer(
 }
 
 func (r *testStatsReporter) ReportTimer(name string, tags map[string]string, interval time.Duration) {
+	r.mtx.Lock()
+	defer r.mtx.Unlock()
+
 	r.timers[name] = &testIntValue{
 		val:  int64(interval),
 		tags: tags,
@@ -290,8 +325,10 @@ type testStatsReporterCachedHistogramValueBucket struct {
 }
 
 func (b testStatsReporterCachedHistogramValueBucket) ReportSamples(v int64) {
-	b.histogram.r.ReportHistogramValueSamples(b.histogram.name, b.histogram.tags,
-		b.histogram.buckets, b.bucketLowerBound, b.bucketUpperBound, v)
+	b.histogram.r.ReportHistogramValueSamples(
+		b.histogram.name, b.histogram.tags,
+		b.histogram.buckets, b.bucketLowerBound, b.bucketUpperBound, v,
+	)
 }
 
 type testStatsReporterCachedHistogramDurationBucket struct {
@@ -301,8 +338,10 @@ type testStatsReporterCachedHistogramDurationBucket struct {
 }
 
 func (b testStatsReporterCachedHistogramDurationBucket) ReportSamples(v int64) {
-	b.histogram.r.ReportHistogramDurationSamples(b.histogram.name, b.histogram.tags,
-		b.histogram.buckets, b.bucketLowerBound, b.bucketUpperBound, v)
+	b.histogram.r.ReportHistogramDurationSamples(
+		b.histogram.name, b.histogram.tags,
+		b.histogram.buckets, b.bucketLowerBound, b.bucketUpperBound, v,
+	)
 }
 
 func (r *testStatsReporter) ReportHistogramValueSamples(
@@ -313,6 +352,9 @@ func (r *testStatsReporter) ReportHistogramValueSamples(
 	bucketUpperBound float64,
 	samples int64,
 ) {
+	r.mtx.Lock()
+	defer r.mtx.Unlock()
+
 	key := KeyForPrefixedStringMap(name, tags)
 	value, ok := r.histograms[key]
 	if !ok {
@@ -332,6 +374,9 @@ func (r *testStatsReporter) ReportHistogramDurationSamples(
 	bucketUpperBound time.Duration,
 	samples int64,
 ) {
+	r.mtx.Lock()
+	defer r.mtx.Unlock()
+
 	key := KeyForPrefixedStringMap(name, tags)
 	value, ok := r.histograms[key]
 	if !ok {
@@ -353,7 +398,7 @@ func (r *testStatsReporter) Flush() {
 
 func TestWriteTimerImmediately(t *testing.T) {
 	r := newTestStatsReporter()
-	s, closer := NewRootScope(ScopeOptions{Reporter: r}, 0)
+	s, closer := NewRootScope(ScopeOptions{Reporter: r, OmitCardinalityMetrics: true}, 0)
 	defer closer.Close()
 	r.tg.Add(1)
 	s.Timer("ticky").Record(time.Millisecond * 175)
@@ -362,7 +407,7 @@ func TestWriteTimerImmediately(t *testing.T) {
 
 func TestWriteTimerClosureImmediately(t *testing.T) {
 	r := newTestStatsReporter()
-	s, closer := NewRootScope(ScopeOptions{Reporter: r}, 0)
+	s, closer := NewRootScope(ScopeOptions{Reporter: r, OmitCardinalityMetrics: true}, 0)
 	defer closer.Close()
 	r.tg.Add(1)
 	tm := s.Timer("ticky")
@@ -372,7 +417,27 @@ func TestWriteTimerClosureImmediately(t *testing.T) {
 
 func TestWriteReportLoop(t *testing.T) {
 	r := newTestStatsReporter()
-	s, closer := NewRootScope(ScopeOptions{Reporter: r}, 10)
+	s, closer := NewRootScope(ScopeOptions{Reporter: r, OmitCardinalityMetrics: true}, 10)
+	defer closer.Close()
+
+	r.cg.Add(1)
+	s.Counter("bar").Inc(1)
+	r.gg.Add(1)
+	s.Gauge("zed").Update(1)
+	r.tg.Add(1)
+	s.Timer("ticky").Record(time.Millisecond * 175)
+	r.hg.Add(1)
+	s.Histogram("baz", MustMakeLinearValueBuckets(0, 10, 10)).
+		RecordValue(42.42)
+
+	r.WaitAll()
+}
+
+func TestWriteReportLoopDefaultInterval(t *testing.T) {
+	r := newTestStatsReporter()
+	s, closer := NewRootScopeWithDefaultInterval(
+		ScopeOptions{Reporter: r, OmitCardinalityMetrics: true},
+	)
 	defer closer.Close()
 
 	r.cg.Add(1)
@@ -390,7 +455,7 @@ func TestWriteReportLoop(t *testing.T) {
 
 func TestCachedReportLoop(t *testing.T) {
 	r := newTestStatsReporter()
-	s, closer := NewRootScope(ScopeOptions{CachedReporter: r}, 10)
+	s, closer := NewRootScope(ScopeOptions{CachedReporter: r, OmitCardinalityMetrics: true}, 10)
 	defer closer.Close()
 
 	r.cg.Add(1)
@@ -408,9 +473,9 @@ func TestCachedReportLoop(t *testing.T) {
 func testReportLoopFlushOnce(t *testing.T, cached bool) {
 	r := newTestStatsReporter()
 
-	scopeOpts := ScopeOptions{CachedReporter: r}
+	scopeOpts := ScopeOptions{CachedReporter: r, OmitCardinalityMetrics: true}
 	if !cached {
-		scopeOpts = ScopeOptions{Reporter: r}
+		scopeOpts = ScopeOptions{Reporter: r, OmitCardinalityMetrics: true}
 	}
 
 	s, closer := NewRootScope(scopeOpts, 10*time.Minute)
@@ -448,7 +513,7 @@ func TestReporterFlushOnce(t *testing.T) {
 func TestWriteOnce(t *testing.T) {
 	r := newTestStatsReporter()
 
-	root, closer := NewRootScope(ScopeOptions{Reporter: r}, 0)
+	root, closer := NewRootScope(ScopeOptions{Reporter: r, OmitCardinalityMetrics: true}, 0)
 	defer closer.Close()
 
 	s := root.(*scope)
@@ -519,9 +584,10 @@ func TestHistogramSharedBucketMetrics(t *testing.T) {
 	var (
 		r     = newTestStatsReporter()
 		scope = newRootScope(ScopeOptions{
-			Prefix:         "",
-			Tags:           nil,
-			CachedReporter: r,
+			Prefix:                 "",
+			Tags:                   nil,
+			CachedReporter:         r,
+			OmitCardinalityMetrics: true,
 		}, 0)
 		builder = func(s Scope) func(map[string]string) {
 			buckets := MustMakeLinearValueBuckets(10, 10, 3)
@@ -544,9 +610,11 @@ func TestHistogramSharedBucketMetrics(t *testing.T) {
 			defer wg.Done()
 
 			val := strconv.Itoa(i % 4)
-			record(map[string]string{
-				"key": val,
-			})
+			record(
+				map[string]string{
+					"key": val,
+				},
+			)
 
 			time.Sleep(time.Duration(rand.Float64() * float64(time.Second)))
 		}()
@@ -580,12 +648,68 @@ func TestHistogramSharedBucketMetrics(t *testing.T) {
 	require.Equal(t, 0, len(unseen), fmt.Sprintf("%v", unseen))
 }
 
+func TestConcurrentUpdates(t *testing.T) {
+	var (
+		r                = newTestStatsReporter()
+		wg               = &sync.WaitGroup{}
+		workerCount      = 20
+		scopeCount       = 4
+		countersPerScope = 4
+		counterIncrs     = 5000
+		rs               = newRootScope(
+			ScopeOptions{
+				Prefix:                 "",
+				Tags:                   nil,
+				CachedReporter:         r,
+				OmitCardinalityMetrics: true,
+			}, 0,
+		)
+		scopes   = []Scope{rs}
+		counters []Counter
+	)
+
+	// Instantiate Subscopes.
+	for i := 1; i < scopeCount; i++ {
+		scopes = append(scopes, rs.SubScope(fmt.Sprintf("subscope_%d", i)))
+	}
+
+	// Instantiate Counters.
+	for sNum, s := range scopes {
+		for cNum := 0; cNum < countersPerScope; cNum++ {
+			counters = append(counters, s.Counter(fmt.Sprintf("scope_%d_counter_%d", sNum, cNum)))
+		}
+	}
+
+	// Instantiate workers.
+	r.cg.Add(scopeCount * countersPerScope)
+	for worker := 0; worker < workerCount; worker++ {
+		wg.Add(1)
+		go func() {
+			defer wg.Done()
+			// Counter should have counterIncrs * workerCount.
+			for i := 0; i < counterIncrs*len(counters); i++ {
+				counters[i%len(counters)].Inc(1)
+			}
+		}()
+	}
+
+	wg.Wait()
+	rs.reportRegistry()
+	r.WaitAll()
+
+	wantVal := int64(workerCount * counterIncrs)
+	for _, gotCounter := range r.getCounters() {
+		assert.Equal(t, gotCounter.val, wantVal)
+	}
+}
+
 func TestCounterSanitized(t *testing.T) {
 	r := newTestStatsReporter()
 
 	root, closer := NewRootScope(ScopeOptions{
-		Reporter:        r,
-		SanitizeOptions: &alphanumericSanitizerOpts,
+		Reporter:               r,
+		SanitizeOptions:        &alphanumericSanitizerOpts,
+		OmitCardinalityMetrics: true,
 	}, 0)
 	defer closer.Close()
 
@@ -641,7 +765,7 @@ func TestCounterSanitized(t *testing.T) {
 func TestCachedReporter(t *testing.T) {
 	r := newTestStatsReporter()
 
-	root, closer := NewRootScope(ScopeOptions{CachedReporter: r}, 0)
+	root, closer := NewRootScope(ScopeOptions{CachedReporter: r, OmitCardinalityMetrics: true}, 0)
 	defer closer.Close()
 
 	s := root.(*scope)
@@ -678,7 +802,7 @@ func TestCachedReporter(t *testing.T) {
 func TestRootScopeWithoutPrefix(t *testing.T) {
 	r := newTestStatsReporter()
 
-	root, closer := NewRootScope(ScopeOptions{Reporter: r}, 0)
+	root, closer := NewRootScope(ScopeOptions{Reporter: r, OmitCardinalityMetrics: true}, 0)
 	defer closer.Close()
 
 	s := root.(*scope)
@@ -712,7 +836,9 @@ func TestRootScopeWithoutPrefix(t *testing.T) {
 func TestRootScopeWithPrefix(t *testing.T) {
 	r := newTestStatsReporter()
 
-	root, closer := NewRootScope(ScopeOptions{Prefix: "foo", Reporter: r}, 0)
+	root, closer := NewRootScope(
+		ScopeOptions{Prefix: "foo", Reporter: r, OmitCardinalityMetrics: true}, 0,
+	)
 	defer closer.Close()
 
 	s := root.(*scope)
@@ -746,7 +872,11 @@ func TestRootScopeWithPrefix(t *testing.T) {
 func TestRootScopeWithDifferentSeparator(t *testing.T) {
 	r := newTestStatsReporter()
 
-	root, closer := NewRootScope(ScopeOptions{Prefix: "foo", Separator: "_", Reporter: r}, 0)
+	root, closer := NewRootScope(
+		ScopeOptions{
+			Prefix: "foo", Separator: "_", Reporter: r, OmitCardinalityMetrics: true,
+		}, 0,
+	)
 	defer closer.Close()
 
 	s := root.(*scope)
@@ -780,7 +910,9 @@ func TestRootScopeWithDifferentSeparator(t *testing.T) {
 func TestSubScope(t *testing.T) {
 	r := newTestStatsReporter()
 
-	root, closer := NewRootScope(ScopeOptions{Prefix: "foo", Reporter: r}, 0)
+	root, closer := NewRootScope(
+		ScopeOptions{Prefix: "foo", Reporter: r, OmitCardinalityMetrics: true}, 0,
+	)
 	defer closer.Close()
 
 	tags := map[string]string{"foo": "bar"}
@@ -822,7 +954,7 @@ func TestSubScope(t *testing.T) {
 func TestSubScopeClose(t *testing.T) {
 	r := newTestStatsReporter()
 
-	rs, closer := NewRootScope(ScopeOptions{Prefix: "foo", Reporter: r}, 0)
+	rs, closer := NewRootScope(ScopeOptions{Prefix: "foo", Reporter: r, OmitCardinalityMetrics: true}, 0)
 	// defer closer.Close()
 	_ = closer
 
@@ -913,7 +1045,11 @@ func TestTaggedSubScope(t *testing.T) {
 	r := newTestStatsReporter()
 
 	ts := map[string]string{"env": "test"}
-	root, closer := NewRootScope(ScopeOptions{Prefix: "foo", Tags: ts, Reporter: r}, 0)
+	root, closer := NewRootScope(
+		ScopeOptions{
+			Prefix: "foo", Tags: ts, Reporter: r, OmitCardinalityMetrics: true,
+		}, 0,
+	)
 	defer closer.Close()
 
 	s := root.(*scope)
@@ -945,19 +1081,23 @@ func TestTaggedSubScope(t *testing.T) {
 	assert.EqualValues(t, ts, counters["foo.beep"].tags)
 
 	assert.EqualValues(t, 1, counters["foo.boop"].val)
-	assert.EqualValues(t, map[string]string{
-		"env":     "test",
-		"service": "test",
-	}, counters["foo.boop"].tags)
+	assert.EqualValues(
+		t, map[string]string{
+			"env":     "test",
+			"service": "test",
+		}, counters["foo.boop"].tags,
+	)
 
 	assert.EqualValues(t, 1, histograms["foo.baz"].valueSamples[50.0])
 	assert.EqualValues(t, ts, histograms["foo.baz"].tags)
 
 	assert.EqualValues(t, 1, histograms["foo.bar"].valueSamples[50.0])
-	assert.EqualValues(t, map[string]string{
-		"env":     "test",
-		"service": "test",
-	}, histograms["foo.bar"].tags)
+	assert.EqualValues(
+		t, map[string]string{
+			"env":     "test",
+			"service": "test",
+		}, histograms["foo.bar"].tags,
+	)
 }
 
 func TestTaggedSanitizedSubScope(t *testing.T) {
@@ -965,13 +1105,13 @@ func TestTaggedSanitizedSubScope(t *testing.T) {
 
 	ts := map[string]string{"env": "test:env"}
 	root, closer := NewRootScope(ScopeOptions{
-		Prefix:          "foo",
-		Tags:            ts,
-		Reporter:        r,
-		SanitizeOptions: &alphanumericSanitizerOpts,
+		Prefix:                 "foo",
+		Tags:                   ts,
+		Reporter:               r,
+		SanitizeOptions:        &alphanumericSanitizerOpts,
+		OmitCardinalityMetrics: true,
 	}, 0)
 	defer closer.Close()
-
 	s := root.(*scope)
 
 	tscope := root.Tagged(map[string]string{"service": "test.service"}).(*scope)
@@ -985,10 +1125,12 @@ func TestTaggedSanitizedSubScope(t *testing.T) {
 
 	counters := r.getCounters()
 	assert.EqualValues(t, 1, counters["foo_beep"].val)
-	assert.EqualValues(t, map[string]string{
-		"env":     "test_env",
-		"service": "test_service",
-	}, counters["foo_beep"].tags)
+	assert.EqualValues(
+		t, map[string]string{
+			"env":     "test_env",
+			"service": "test_service",
+		}, counters["foo_beep"].tags,
+	)
 }
 
 func TestTaggedExistingReturnsSameScope(t *testing.T) {
@@ -998,7 +1140,11 @@ func TestTaggedExistingReturnsSameScope(t *testing.T) {
 		nil,
 		{"env": "test"},
 	} {
-		root, closer := NewRootScope(ScopeOptions{Prefix: "foo", Tags: initialTags, Reporter: r}, 0)
+		root, closer := NewRootScope(
+			ScopeOptions{
+				Prefix: "foo", Tags: initialTags, Reporter: r, OmitCardinalityMetrics: true,
+			}, 0,
+		)
 		defer closer.Close()
 
 		rootScope := root.(*scope)
@@ -1018,7 +1164,6 @@ func TestSnapshot(t *testing.T) {
 	commonTags := map[string]string{"env": "test"}
 	s := NewTestScope("foo", map[string]string{"env": "test"})
 	child := s.Tagged(map[string]string{"service": "test"})
-
 	s.Counter("beep").Inc(1)
 	s.Gauge("bzzt").Update(2)
 	s.Timer("brrr").Record(1 * time.Second)
@@ -1030,52 +1175,94 @@ func TestSnapshot(t *testing.T) {
 
 	// Should be able to call Snapshot any number of times and get same result.
 	for i := 0; i < 3; i++ {
-		t.Run(fmt.Sprintf("attempt %d", i), func(t *testing.T) {
-			snap := s.Snapshot()
-			counters, gauges, timers, histograms :=
-				snap.Counters(), snap.Gauges(), snap.Timers(), snap.Histograms()
+		t.Run(
+			fmt.Sprintf("attempt %d", i), func(t *testing.T) {
+				snap := s.Snapshot()
+				counters, gauges, timers, histograms :=
+					snap.Counters(), snap.Gauges(), snap.Timers(), snap.Histograms()
 
-			assert.EqualValues(t, 1, counters["foo.beep+env=test"].Value())
-			assert.EqualValues(t, commonTags, counters["foo.beep+env=test"].Tags())
+				assert.EqualValues(t, 1, counters["foo.beep+env=test"].Value())
+				assert.EqualValues(t, commonTags, counters["foo.beep+env=test"].Tags())
 
-			assert.EqualValues(t, 2, gauges["foo.bzzt+env=test"].Value())
-			assert.EqualValues(t, commonTags, gauges["foo.bzzt+env=test"].Tags())
+				assert.EqualValues(t, 2, gauges["foo.bzzt+env=test"].Value())
+				assert.EqualValues(t, commonTags, gauges["foo.bzzt+env=test"].Tags())
 
-			assert.EqualValues(t, []time.Duration{
-				1 * time.Second,
-				2 * time.Second,
-			}, timers["foo.brrr+env=test"].Values())
-			assert.EqualValues(t, commonTags, timers["foo.brrr+env=test"].Tags())
+				assert.EqualValues(
+					t, []time.Duration{
+						1 * time.Second,
+						2 * time.Second,
+					}, timers["foo.brrr+env=test"].Values(),
+				)
+				assert.EqualValues(t, commonTags, timers["foo.brrr+env=test"].Tags())
 
-			assert.EqualValues(t, map[float64]int64{
-				0:               0,
-				2:               1,
-				4:               0,
-				math.MaxFloat64: 1,
-			}, histograms["foo.fizz+env=test"].Values())
-			assert.EqualValues(t, map[time.Duration]int64(nil), histograms["foo.fizz+env=test"].Durations())
-			assert.EqualValues(t, commonTags, histograms["foo.fizz+env=test"].Tags())
+				assert.EqualValues(
+					t, map[float64]int64{
+						0:               0,
+						2:               1,
+						4:               0,
+						math.MaxFloat64: 1,
+					}, histograms["foo.fizz+env=test"].Values(),
+				)
+				assert.EqualValues(t, map[time.Duration]int64(nil), histograms["foo.fizz+env=test"].Durations())
+				assert.EqualValues(t, commonTags, histograms["foo.fizz+env=test"].Tags())
 
-			assert.EqualValues(t, map[float64]int64(nil), histograms["foo.buzz+env=test"].Values())
-			assert.EqualValues(t, map[time.Duration]int64{
-				time.Second * 2: 1,
-				time.Second * 4: 0,
-				math.MaxInt64:   0,
-			}, histograms["foo.buzz+env=test"].Durations())
-			assert.EqualValues(t, commonTags, histograms["foo.buzz+env=test"].Tags())
+				assert.EqualValues(t, map[float64]int64(nil), histograms["foo.buzz+env=test"].Values())
+				assert.EqualValues(
+					t, map[time.Duration]int64{
+						time.Second * 2: 1,
+						time.Second * 4: 0,
+						math.MaxInt64:   0,
+					}, histograms["foo.buzz+env=test"].Durations(),
+				)
+				assert.EqualValues(t, commonTags, histograms["foo.buzz+env=test"].Tags())
 
-			assert.EqualValues(t, 1, counters["foo.boop+env=test,service=test"].Value())
-			assert.EqualValues(t, map[string]string{
-				"env":     "test",
-				"service": "test",
-			}, counters["foo.boop+env=test,service=test"].Tags())
-		})
+				assert.EqualValues(t, 1, counters["foo.boop+env=test,service=test"].Value())
+				assert.EqualValues(
+					t, map[string]string{
+						"env":     "test",
+						"service": "test",
+					}, counters["foo.boop+env=test,service=test"].Tags(),
+				)
+			},
+		)
 	}
 }
 
+func TestSnapshotConcurrent(t *testing.T) {
+	var (
+		scope = NewTestScope("", nil)
+		quit  = make(chan struct{})
+		done  = make(chan struct{})
+	)
+
+	go func() {
+		defer close(done)
+		for {
+			select {
+			case <-quit:
+				return
+			default:
+				hello := scope.Tagged(map[string]string{"a": "b"}).Counter("hello")
+				hello.Inc(1)
+			}
+		}
+	}()
+	var val CounterSnapshot
+	for {
+		val = scope.Snapshot().Counters()["hello+a=b"]
+		if val != nil {
+			quit <- struct{}{}
+			break
+		}
+	}
+	require.NotNil(t, val)
+
+	<-done
+}
+
 func TestCapabilities(t *testing.T) {
 	r := newTestStatsReporter()
-	s, closer := NewRootScope(ScopeOptions{Reporter: r}, 0)
+	s, closer := NewRootScope(ScopeOptions{Reporter: r, OmitCardinalityMetrics: true}, 0)
 	defer closer.Close()
 	assert.True(t, s.Capabilities().Reporting())
 	assert.False(t, s.Capabilities().Tagging())
@@ -1103,7 +1290,8 @@ func TestScopeDefaultBuckets(t *testing.T) {
 			90 * time.Millisecond,
 			120 * time.Millisecond,
 		},
-		Reporter: r,
+		Reporter:               r,
+		OmitCardinalityMetrics: true,
 	}, 0)
 	defer closer.Close()
 
@@ -1134,7 +1322,7 @@ func newTestMets(scope Scope) testMets {
 func TestReturnByValue(t *testing.T) {
 	r := newTestStatsReporter()
 
-	root, closer := NewRootScope(ScopeOptions{Reporter: r}, 0)
+	root, closer := NewRootScope(ScopeOptions{Reporter: r, OmitCardinalityMetrics: true}, 0)
 	defer closer.Close()
 
 	s := root.(*scope)
@@ -1151,7 +1339,7 @@ func TestReturnByValue(t *testing.T) {
 
 func TestScopeAvoidReportLoopRunOnClose(t *testing.T) {
 	r := newTestStatsReporter()
-	root, closer := NewRootScope(ScopeOptions{Reporter: r}, 0)
+	root, closer := NewRootScope(ScopeOptions{Reporter: r, OmitCardinalityMetrics: true}, 0)
 
 	s := root.(*scope)
 	s.reportLoopRun()
@@ -1166,7 +1354,7 @@ func TestScopeAvoidReportLoopRunOnClose(t *testing.T) {
 
 func TestScopeFlushOnClose(t *testing.T) {
 	r := newTestStatsReporter()
-	root, closer := NewRootScope(ScopeOptions{Reporter: r}, time.Hour)
+	root, closer := NewRootScope(ScopeOptions{Reporter: r, OmitCardinalityMetrics: true}, time.Hour)
 
 	r.cg.Add(1)
 	root.Counter("foo").Inc(1)
